package DistroDB::BinaryPremap;

use strict;
use warnings;

use DistroDB::CLI::DistroMap qw/$opt_use_projectdb/;
use DistroDB::CLI::DebugName qw/$debug_obname $debug_oname/;
use DistroDB::BinaryDistroMap qw/&add_distromap/;
use DistroDB::Pkgset qw/
&__pkgset_get_pkgarrayref
&__pkgset_pkgs
&__pkgset_power
/;
use DistroDB::Arrow qw/
&static_create_raw_arrow
&__sort_arrowset_by_fingerprint_weight
&__aggregate_arrowset_by_pkgset
&__get_arrow_keystring
&__arrowset_debug
&__arrowset_split_by_connected_components
&__arrow_head_pkgs
&__arrow_get_head_pkgset
&__arrow_get_fingerprint
/;
use DistroDB::Fingerprint qw/
&__fingerprint_weight
&__fingerprint_is_empty
&__fingerprint_intersect_and_setminus
/;
use ProjectDBTools;
use DistroDB::Tags qw/@optional_binary_tags_order/;
use DistroDB::OriginDestTags qw/
$origin_SST_db
$origin_BOT_db
$origin_BMT_db
$origin_BCT_db
$dest_SST_db
$dest_BOT_db
$dest_BMT_db
$dest_BCT_db
/;

our $verbose=0;

#my vars for package DistroDB::BinaryPackageMap;
my $debug_reject_SST=0;
# TODO: exclude_me_from_unmapped in source-distrodb2distromap
our %originsrc_reject_SST;
# save as noversion flag as same names are mapped by default
our %originsrc_selfreject_SST;

my ($originurl,$desturl,$originsourcename,$destsourcename);

sub initialize {
    my (undef,$origin,$dest)=@_;
    $originurl=$origin->{'sourceurl.raw'};
    $desturl=$dest->{'sourceurl.raw'};
    $originsourcename=$origin->{'sourcename'};
    $destsourcename=$dest->{'sourcename'};
}

sub new {
    my $class=shift;
    my $self={
	@_,
    };
    bless $self, $class;

    my $locdebug;
    my $originpkg=$self->{-originpkg};
    if ($verbose>2
	or $debug_obname and $debug_obname eq $originpkg
	or $debug_oname and $debug_oname eq $self->get_originsrc()
	) {
	$locdebug=1;
	$locdebug=$verbose if $verbose;
	$self->{-debug}=$locdebug;
    }
    print STDERR "originpkg=$originpkg starting...\n" if $locdebug;

    return $self;
}

sub get_ordered_origin_BOT_listref {
    my $self=shift;
    my $BOT=$self->{-origin_BOT_list};
    return $BOT if $BOT;
    my $originpkg=$self->{-originpkg};
    my $origin_BOT=$origin_BOT_db->{$originpkg};
    my @origin_BOT_list=&_get_ordered_BOT_list($origin_BOT);
    $BOT=\@origin_BOT_list;
    $self->{-origin_BOT_list}=$BOT;
    return $BOT;
}

sub get_origin_BMT_hashref {
    my $self=shift;
    my $BMT=$self->{-origin_BMT_hash};
    return $BMT if $BMT;
    $BMT=$origin_BMT_db->{$self->{-originpkg}};
    $self->{-origin_BMT_hash}=$BMT;
    return $BMT;
}

sub get_origin_BCT_hashref {
    my $self=shift;
    my $BCT=$self->{-origin_BCT_hash};
    return $BCT if $BCT;
    $BCT=$origin_BCT_db->{$self->{-originpkg}};
    $self->{-origin_BCT_hash}=$BCT;
    return $BCT;
}

sub get_originsrc {
    my $self=shift;
    my $originsrc=$self->{-originsrc};
    return $originsrc if $originsrc;
    $originsrc=&_get_originsrc($self->{-originpkg});
    $self->{-originsrc}=$originsrc;
    return $originsrc;
}

# EXTERNAL $originsourcename is here
sub _get_originsrc {
    my $originpkg=shift;
    my $originsrc=$originsourcename->{$originpkg};
    if (!defined($originsrc)) {
	warn "W: can't find origin src for $originpkg" unless defined $originsrc;
	# 30% chance
	$originsrc=$originpkg;
	# will confuse is same url
	#$originsourcename->{$originpkg}=$originsrc;
    }
    return $originsrc;
}

sub premap_originpkg {
    my $self=shift;
    my $originpkg=$self->{-originpkg};
    my $locdebug=$self->{-debug};
    my $untrimmed_arrowset=$self->{-untrimmed_arrowsref};
    if ($locdebug) {
	warn "debug: untrimmed arrowset:\n";
	&__arrowset_debug($untrimmed_arrowset,$originpkg);
    }
    my $arrowset=$self->trim_arrowset_by_SST($untrimmed_arrowset);
    if ($locdebug) {
	warn "debug: trimmed_by_SST arrowset:\n";
	&__arrowset_debug($arrowset,$originpkg);
    }

    my $same_map_found=0;
    my ($same_arrowset,$replaceable_arrowset)=$self->arrowset_split_by_same_url($arrowset);
    my $have_same_dist=scalar @$same_arrowset;
    if ($have_same_dist) {
	my $bmt_trimmed_same_arrowset=$self->trim_arrowset_by_BMT_BCT($same_arrowset);
	$same_map_found=scalar @$bmt_trimmed_same_arrowset;
	$self->arrow_translate_connected_components($bmt_trimmed_same_arrowset);
    } else {
	my $bmt_trimmed_replaceable_arrowset=$self->trim_arrowset_by_BMT_BCT($replaceable_arrowset);
	$self->arrow_translate_connected_components($bmt_trimmed_replaceable_arrowset);
    }
    $self->{-same_map_found}=$same_map_found;
    $self->{-have_same_dist}=$have_same_dist;
}

sub arrow_translate_connected_components {
    my ($self, $arrowset)=@_;
    return unless @$arrowset>0;
    my $locdebug=$self->{-debug};
    print STDERR 'DEBUG: ',$self->{-originpkg},': got ', scalar @$arrowset,' arrows, splitting into connected components ...',"\n" if $locdebug;
    my @connected_components=map {$self->trim_arrowset_by_BOT($_)} &__arrowset_split_by_connected_components($arrowset);
    print STDERR 'DEBUG: got ',scalar @connected_components,' connected components',"\n" if $locdebug;
    foreach my $arrowset (@connected_components) {
	$self->arrow_translate_connected_component($arrowset);
    }
    print STDERR 'DEBUG: translating ... done',"\n" if $locdebug;
}

#
# TODO
#
sub arrow_translate_connected_component0 {
    my ($self, $arrowset)=@_;

    # TODO: fixme! hack, not a proper algorythm
    my $sortedset=&__sort_arrowset_by_fingerprint_weight($arrowset);
    my @collapsed_set=($sortedset->[0]);
    my $fingerprint1=&__arrow_get_fingerprint($sortedset->[0]);
    for (my $i=2;$i<@$sortedset;$i++) {
	my $fingerprint2=&__arrow_get_fingerprint($sortedset->[$i]);
	my ($union,$diff)=&__fingerprint_intersect_and_setminus($fingerprint2,$fingerprint1);
	if (&__fingerprint_is_empty($diff) and &__fingerprint_weight($fingerprint2)/&__fingerprint_weight($fingerprint1)< .3) {
	    # collapse
	} else {
	    push @collapsed_set, $sortedset->[$i];
	}
    }

    return if @collapsed_set>1;
    # TODO: intersect originset and same candidates, get a choice, select one
    foreach my $arrow (@collapsed_set) {
	$self->arrow_translate($arrow);
    }
}

sub arrow_translate_connected_component {
    my ($self, $arrowset)=@_;
    return if @$arrowset>1;
    # TODO: intersect originset and same candidates, get a choice, select one
    foreach my $arrow (@$arrowset) {
	$self->arrow_translate($arrow);
    }
}


# common cache shared with binary subpackages
my $is_SST_cache={};

sub __get_originsrc_cache_is_SST {
    my $originsrc=shift;
    my $originsrc_cache_is_SST=$is_SST_cache->{$originsrc};
    return $originsrc_cache_is_SST if $originsrc_cache_is_SST;
    $originsrc_cache_is_SST={};
    $is_SST_cache->{$originsrc}=$originsrc_cache_is_SST;
    return $originsrc_cache_is_SST;
}

sub trim_arrowset_by_SST {
    my ($self, $arrowset)=@_;
    my $originpkg=$self->{-originpkg};
    my $locdebug=$self->{-debug};
    my $originsrc=$self->get_originsrc();
    my $origin_SST=$origin_SST_db->{$originsrc};
    # for packages with optional support of kdeX
    return $arrowset unless $origin_SST and %$origin_SST;
    my $originsrc_cache_is_SST=&__get_originsrc_cache_is_SST($originsrc);
    my @new_arrows;
    foreach my $arrow (@$arrowset) {
	my $old_destset=&__arrow_get_head_pkgset($arrow);
	my @SST_dest_set=grep {&_the_same_SST($origin_SST,$originsrc,$originsrc_cache_is_SST,$_,$locdebug)} &__pkgset_pkgs($old_destset);
	if (&__pkgset_power($old_destset) == @SST_dest_set) {
	    push @new_arrows, $arrow;
	} elsif (@SST_dest_set>0) {
	    print STDERR "DEBUG: arrow_check_same_url: $originpkg:\n",
		"\torigin_SST=",join(' ',keys(%$origin_SST)),"\n",
		"\tSST_dest_set=",join(' ',@SST_dest_set),"\n",
		if $locdebug;
	    push @new_arrows, &static_create_raw_arrow(&__arrow_get_fingerprint($arrow),\@SST_dest_set);
	}
    }
    return \@new_arrows;
}

sub arrowset_split_by_same_url {
    my ($self, $arrowset)=@_;
    my $originpkg=$self->{-originpkg};
    my $locdebug=$self->{-debug};
    my (@samedest_arrows,@replaceabledest_arrows);
    foreach my $arrow (@$arrowset) {
	my (@same_dest_set,@replaceable_dest_set);
	foreach my $destpkg (&__arrow_head_pkgs($arrow)) {
	    my $is_same=&_is_same_url_origin2dest($originpkg, $destpkg,
		  &__get_arrow_keystring($arrow),$locdebug);
	    print STDERR 'DEBUG: ',&__get_arrow_keystring($arrow),
	    " originpkg=$originpkg destpkg=$destpkg sameurl=$is_same\n" if $locdebug;
	    push @same_dest_set, $destpkg if $is_same == ProjectDBTools::SAME or $is_same == ProjectDBTools::SUBPROJECT;
	    push @replaceable_dest_set, $destpkg if $is_same== ProjectDBTools::REPLACEABLE;
	}
	print STDERR "DEBUG: arrow_check_same_url: $originpkg:\n",
	    "\tsame_dest_set=",join(' ',@same_dest_set),"\n",
	    "\treplaceable_dest_set=",join(' ',@replaceable_dest_set),"\n",
	    if $locdebug;
	if (@same_dest_set) {
	    push @samedest_arrows, &static_create_raw_arrow(&__arrow_get_fingerprint($arrow),\@same_dest_set);
	} elsif (@replaceable_dest_set) {
	    push @replaceabledest_arrows, &static_create_raw_arrow(&__arrow_get_fingerprint($arrow),\@replaceable_dest_set);
	}
    }
    return
	&__aggregate_arrowset_by_pkgset(\@samedest_arrows),
	&__aggregate_arrowset_by_pkgset(\@replaceabledest_arrows);
}

sub trim_arrowset_by_BMT_BCT {
    my ($self, $arrowset)=@_;
    my $originpkg=$self->{-originpkg};
    my $locdebug=$self->{-debug};
    my @new_arrows;
    foreach my $arrow (@$arrowset) {
	my $old_destset=&__arrow_get_head_pkgset($arrow);
	my $old_destsetref=&__pkgset_get_pkgarrayref($old_destset);
	my $destsetref=$old_destsetref;
	# BMT tag filtering
	my @filered_BMT=&_filter_by_BMT($self->get_origin_BMT_hashref(),$old_destsetref,$locdebug);
	if (&__pkgset_power($old_destset) != @filered_BMT) {
	    $destsetref=\@filered_BMT;
	    print STDERR "\tdestset=",join(' ',@$destsetref),"\n",
		"\tfiltered_BMT=",join(' ',@filered_BMT),"\n" if $locdebug;
	}
	# BCT tag filtering
	my @filered_BCT=&_filter_by_BCT($self->get_origin_BCT_hashref(),\@filered_BMT,$locdebug);
	if ($#filered_BCT != $#filered_BMT) {
	    $destsetref=\@filered_BCT;
	    print STDERR "\tdestset=",join(' ',@$destsetref),"\n",
		"\tfiltered_BCT=",join(' ',@filered_BCT),"\n" if $locdebug;
	}
	if (&__pkgset_power($old_destset)== scalar @$destsetref) {
	    push @new_arrows, $arrow;
	} elsif (@$destsetref==0) {
	    # skip arrow; nothing to map to
	} else {
	    push @new_arrows, &static_create_raw_arrow(&__arrow_get_fingerprint($arrow),$destsetref);
	}
    }
    return &__aggregate_arrowset_by_pkgset(\@new_arrows);
}

sub trim_arrowset_by_BOT {
    my ($self, $arrowset)=@_;
    my $locdebug=$self->{-debug};
    my @old_destset=map {&__arrow_head_pkgs($_)} @$arrowset;
    if (@old_destset < 2) {
	print STDERR 'trim_arrowset_by_BOT: skipped',"\n" if $locdebug;
	return $arrowset;
    }
    print STDERR 'DEBUG: trim_arrowset_by_BOT: ',$self->{-originpkg}, ":\n",
	"\tdestset=",join(' ',@old_destset),"\n",
	"\torigin_BOT=",join(' ',@{$self->get_ordered_origin_BOT_listref()}),"\n",
	if $locdebug;
    # tag filtering
    my @filered_destset=&_filter_by_BOT($self->get_ordered_origin_BOT_listref(),\@old_destset,$locdebug);
    if (@filered_destset<1 or scalar @filered_destset == scalar @old_destset) {
	print STDERR 'trim_arrowset_by_BOT: failed to trim',"\n" if $locdebug;
	return $arrowset;
    }
    my %filered_map=map{$_=>1} @filered_destset;
    my @new_arrows;
    foreach my $arrow (@$arrowset) {
	my @old_arrow_destset=&__arrow_head_pkgs($arrow);
	my @filered_arrow_destset=grep {$filered_map{$_}} @old_arrow_destset;
	if (scalar @old_arrow_destset == scalar @filered_arrow_destset) {
	    push @new_arrows, $arrow;
	} elsif (0==@filered_arrow_destset) {
	    # drop
	} else {
	    push @new_arrows, &static_create_raw_arrow(&__arrow_get_fingerprint($arrow),\@filered_arrow_destset);

	}
    }
    return &__aggregate_arrowset_by_pkgset(\@new_arrows);
}

sub arrow_translate {
    my $self=shift;
    my $originpkg=$self->{-originpkg};
    my $locdebug=$self->{-debug};
    my ($arrow)=@_;
    my $destset=&__arrow_get_head_pkgset($arrow);
    my $power=&__pkgset_power($destset);
    if ($power>1) {
	# to do: version corellation, like in
	# libqalculate-devel vs libqalculate4-devel for stem=devel-libs key=qalculate

	# compat filtering
	# vala-compat-tools vs vala-tools for stem=pkg-config key=vapigen
	# flex-devel vs flex-old for stem=static-libs key=l

    } elsif (1==$power) {
	&add_distromap($originpkg,&__pkgset_pkgs($destset));
    } else {
	# todo
    }
}


sub _get_ordered_BOT_list {
    my ($BOT)=@_;
    return unless $BOT;
    my @out;
    foreach my $tag (@optional_binary_tags_order) {
	push @out, $tag if $BOT->{$tag};
    }
    return @out;
}

sub _the_same_SST {
    my ($origin_SST,$originsrc,$origin_cache_is_SST,$destpkg,$locdebug)=@_;
    my $destsrc=$destsourcename->{$destpkg};
    if (!defined($destsrc)) {
	warn "W: can't find dest src for $destpkg" unless defined $destsrc;
	# 30% chance
	$destsrc=$destpkg;
    }
    my $destsrc_is_SST=$origin_cache_is_SST->{$destsrc};
    if (defined $destsrc_is_SST) {
	return $destsrc_is_SST;
    } else {
	$destsrc_is_SST=&__is_the_same_SST($origin_SST,$originsrc,$destsrc,$locdebug);
	$origin_cache_is_SST->{$destsrc}=$destsrc_is_SST;
	return $destsrc_is_SST;
    }
}

sub __is_the_same_SST {
    my ($origin_SST,$originsrc,$destsrc,$locdebug)=@_;
    my $dest_SST=$dest_SST_db->{$destsrc};
    print STDERR "DEBUG: destsrc $destsrc: dest_SST=",join(' ',keys(%$dest_SST)),"\n" if $locdebug;
    my $origin_power=scalar keys(%$origin_SST);
    my $dest_power=scalar keys(%$dest_SST);
    return 1 if $origin_power==0 or $dest_power==0;
    #return 0 if $origin_power != $dest_power;
    return 1 if
	&__keys_are_subset($dest_SST,$origin_SST) or
	&__keys_are_subset($origin_SST,$dest_SST);
    $originsrc_reject_SST{$originsrc}=1;
    # same name in origin and dest;
    # should not appear on watch.altlinux.org -- SST environments are different
    if ($destsrc eq $originsrc) {
	$originsrc_selfreject_SST{$originsrc}=1;
    }
    # TODO: save to a todo file?
    if ($debug_reject_SST or $locdebug) {
	print $originsrc,"\t",join(",",keys(%$origin_SST));
	print "\t", $destsrc, "\t", join(",",keys(%{$dest_SST_db->{$destsrc}}));
	print "\n";
    }
    return 0;
}

sub __keys_are_subset {
    my ($hashref1, $hashref2)=@_;
    foreach my $tag (keys(%$hashref1)) {
	return 0 unless $hashref2->{$tag};
    }
    return 1;
}

sub _filter_by_BMT {
    my ($origin_BMT_hash,$destset,$locdebug)=@_;
    my @oldlist=@$destset;
    my @new_BMT_list;
    if ($origin_BMT_hash and %$origin_BMT_hash) {
        BOM: foreach my $destpkg (@oldlist) {
	    my $dest_BMT=$dest_BMT_db->{$destpkg};
	    next BOM if ! $dest_BMT;
	    foreach my $tag (keys (%$origin_BMT_hash)) {
		next BOM if ! $dest_BMT->{$tag};
		foreach my $dest_tag (keys(%$dest_BMT)) {
		    next BOM if ! $origin_BMT_hash->{$dest_tag};
		}
	    }
	    push @new_BMT_list, $destpkg;
	}
	print STDERR "DEBUG: _filter_by_BMT: new list=",join(' ',@new_BMT_list),"\n" if $locdebug;
    } else {
	foreach my $destpkg (@oldlist) {
	    my $dest_BMT=$dest_BMT_db->{$destpkg};
	    push @new_BMT_list, $destpkg if !$dest_BMT or !%$dest_BMT;
	}
	print STDERR "DEBUG: _filter_by_BMT: new list=",join(' ',@new_BMT_list),"\n" if $locdebug;
    }
    return @new_BMT_list;
}

sub _filter_by_BOT {
    my ($origin_BOT_list,$destset,$locdebug)=@_;
    my @oldlist=@$destset;
    my @newlist;
    foreach my $tag (@$origin_BOT_list) {
	@newlist=();
	foreach my $destpkg (@oldlist) {
	    if ($dest_BOT_db->{$destpkg}->{$tag}) {
		push @newlist, $destpkg;
	    }
	}
	@oldlist=@newlist;
	last if @newlist==1;
    }
    print STDERR 'DEBUG: _filter_by_BOT: new list=',join(' ',@newlist),"\n" if $locdebug;
    return @newlist;
}

sub _filter_by_BCT {
    my ($origin_BCT_set,$destset,$locdebug)=@_;
    return @$destset if not $origin_BCT_set;
    my $origin_BCT_tagset_uuid=join(',',sort {$a cmp $b} keys(%$origin_BCT_set));
    my @newlist;
    foreach my $destpkg (@$destset) {
	my $dest_BCT_set=$dest_BCT_db->{$destpkg};
	unless ($dest_BCT_set) {
	    push @newlist, $destpkg;
	    print STDERR 'DEBUG: BCT: accepted (no dest BCT)',$destpkg,"\n" if $locdebug;
	} else{
	    my $dest_BCT_tagset_uuid=join(',',sort {$a cmp $b} keys(%$dest_BCT_set));
	    #_is_same_set()
	    if ($origin_BCT_tagset_uuid eq $dest_BCT_tagset_uuid) {
		push @newlist, $destpkg;
		print STDERR 'DEBUG: BCT: accepted ',$destpkg,' dest BCT=',$dest_BCT_tagset_uuid,' origin_BCT=',$origin_BCT_tagset_uuid,"\n" if $locdebug;
	    } else {
		# tag conflict
		print STDERR 'DEBUG: BCT: rejected ',$destpkg,' dest BCT=',$dest_BCT_tagset_uuid,' origin_BCT=',$origin_BCT_tagset_uuid,"\n" if $locdebug;
	    }
	}
    }
    print STDERR 'DEBUG: _filter_by_BCT: new list=',join(' ',@newlist),"\n" if $locdebug;
    return @newlist;
}

sub _is_same_url_origin2dest {
    my ($originpkg,$destpkg,$message,$locdebug)=@_;
    my $same_proj=ProjectDBTools::SAME;
    if ($opt_use_projectdb) {
	my $originsrc=$originsourcename->{$originpkg};
	my $destsrc=$destsourcename->{$destpkg};
	if (!defined($originsrc) or !defined($destsrc)) {
	    warn "W: can't find origin src for $originpkg" unless defined $originsrc;
	    warn "W: can't find dest src for $destpkg" unless defined $destsrc;
	} else {
	    $message//='[?]';
	    $same_proj=&ProjectDBTools::origin2dest_same_url(
		$originsrc,$destsrc,$originurl->{$originsrc}, $desturl->{$destsrc},
		"$originpkg vs $destpkg for:\n$message",$locdebug);
	}
    }
    return $same_proj;
}

1;
