package DistroDBTools;

use strict;
use warnings;
use Source::Shared::FindLocalMirror::ALTLinux;

our $verbose=0;

my $distrodbprefix=$ENV{'HOME'}.'/.cache/distrodb';

our @BINARY_DISTRODB_GENERATED=qw!
binary/10-distrodb-old.txt
binary/40-distrodb.txt
binary/40-distrodb-new.txt
!;
our @SOURCE_DISTRODB_GENERATED=qw!
source/10-distrodb-old.txt
source/40-distrodb.txt
source/40-distrodb-new.txt
!;
our %BINARY_DISTRODB_GENERATED=map{$_ => 1} @BINARY_DISTRODB_GENERATED;
our %SOURCE_DISTRODB_GENERATED=map{$_ => 1} @SOURCE_DISTRODB_GENERATED;

sub read_distromap_config {
    die 'not a distromap catalog' if ! -d '.distrodb' and ! $ENV{RPM_BUILD_ROOT};
    my $config={
	DESTREPO => 'altlinux',
	DESTBRANCH => 'sisyphus',
	USE_PROJECTDB => 1,
    };
    open (my $fn, '<', '.distrodb/config') || die "open .distrodb/config: $!";
    local $_;
    while (<$fn>) {
	chomp;
	s/\s*$//;
	$config->{$1}=$2 if /^(\S+)=["']?([^"']+)["']?$/;
    }
    close($fn) || die "close .distrodb/config: $!";

    unless ($config->{'DEST_SRPMS'}) {
	my @args=(-subpath=>'files/SRPMS',-branch=>'Sisyphus');
	my @ALT=&Source::Shared::FindLocalMirror::ALTLinux::find(@args);
	my @AUTOIMPORTS=&Source::Shared::FindLocalMirror::ALTLinux::find(@args,-overlay=>'autoimports');
	die "distromap_config: DEST_SRPMS not set and ALT Linux Sisyphus mirror not found.\n" if ! @ALT;
	warn "distromap_config: DEST_SRPMS not set and ALTLinux autoimports/Sisyphus mirror not found. Map will be incomplete\n" if ! @AUTOIMPORTS;
	$config->{'DEST_SRPMS'} = $ALT[0].(@AUTOIMPORTS>0 ? ' '.$AUTOIMPORTS[0] : '');
    }
    return $config;
}


sub read_distrodb_with {
    my ($map,$distro,$dbname,$func,@funcargs)=@_;
    foreach my $dbfile (glob $distrodbprefix.'/'.$distro.'/'.$dbname.'/*.txt') {
	print STDERR "reading $dbfile...\n" if $verbose;
	&$func($map, $dbfile,@funcargs);
    }
}

sub read_distrodb_map {
    &read_distrodb_with(@_[0..2],\&read_map);
}
sub read_distrodb_tsv_map_overvrite {
    &read_distrodb_with(@_[0..2],\&read_tsv_map);
}
sub read_distrodb_tsv_map_merge {
    &read_distrodb_with(@_[0..2],\&read_tsv_map_merge);
}
sub read_distrodb_tsv_map_sorted {
    &read_distrodb_with(@_[0..2],\&read_tsv_map_sorted);
}
sub read_distrodb_flag {
    &read_distrodb_with(@_[0..2],\&read_flag);
}
sub read_distrodb_2nd_column_as_flag {
    &read_distrodb_with(@_[0..2],\&read_2nd_column_as_flag);
}
sub read_distrodb_one2one {
    &read_distrodb_with(@_[0..2],\&read_one2one);
}
sub read_distrodb_first_one2one {
    &read_distrodb_with(@_[0..2],\&read_first_one2one);
}
sub read_distrodb_one2one_with_resolver_table {
    &read_distrodb_with(@_[0..2],\&read_one2one_with_resolver_table,$_[3]);
}
sub read_distrodb_reverse_known_one2one {
    &read_distrodb_with(@_[0..2],\&read_reverse_known_one2one);
}
sub read_distrodb_stringval {
    &read_distrodb_with(@_[0..2],\&read_stringval);
}

sub read_map {
    my ($map,$filename)=@_;
    open (my $fn,'<',$filename) or die "can't open $filename: $!";
    while (<$fn>) {
	chomp;
	next if /^\s*#/;
	my @in=split(/\s+/);
	next unless @in;
	$map->{shift @in}=\@in;
    }
    close($fn);
}

sub read_tsv_map_overvrite {
    my ($map,$filename)=@_;
    open (my $fn,'<',$filename) or die "can't open $filename: $!";
    while (<$fn>) {
	chomp;
	next if /^\s*#/;
	my @in=split(/\t+/);
	next unless @in;
	$map->{shift @in}=\@in;
    }
    close($fn);
}

sub read_tsv_map_merge {
    my ($map,$filename)=@_;
    open (my $fn,'<',$filename) or die "can't open $filename: $!";
    while (<$fn>) {
	chomp;
	next if /^\s*#/;
	my @in=split(/\t+/);
	next unless @in;
	push @{$map->{shift @in}}, @in;
    }
    close($fn);
}

sub read_tsv_map_sorted {
    my ($map,$filename)=@_;
    open (my $fn,'<',$filename) or die "can't open $filename: $!";
    while (<$fn>) {
	chomp;
	next if /^\s*#/;
	my @in=split(/\t+/);
	next unless @in;
	my $key=shift @in;
	$map->{$key}=[sort {$a cmp $b} @in];
    }
    close($fn);
}

sub write_map {
    my ($map,$filename)=@_;
    open (my $fn,'>',$filename) or die "can't write $filename: $!";
    foreach my $key (sort {$a cmp $b} keys(%$map)) {
	my @val=@{$map->{$key}};
	print $fn join("\t",$key, @val),"\n";
    }
    close($fn);
}

sub write_set {
    my ($set,$filename)=@_;
    open (my $fn,'>',$filename) or die "can't write $filename: $!";
    foreach my $key (sort {$a cmp $b} keys(%$set)) {
	my $val=$set->{$key};
	print $fn join("\t",$key, sort {$a cmp $b} keys(%$val)),"\n";
    }
    close($fn);
}

sub write_one2one {
    my ($map,$filename)=@_;
    open (my $fn,'>',$filename) or die "can't write $filename: $!";
    foreach my $key (sort {$a cmp $b} keys(%$map)) {
	print $fn $key, "\t", $map->{$key}, "\n";
    }
    close($fn);
}

sub write_flag {
    my ($map,$filename)=@_;
    open (my $fn,'>',$filename) or die "can't write $filename: $!";
    foreach my $key (sort {$a cmp $b} keys(%$map)) {
	print $fn $key, "\n";
    }
    close($fn);
}

sub read_flag {
    my ($map,$filename)=@_;
    open (my $fn,'<',$filename) or die "$!";
    while (<$fn>) {
	chomp;
	next if /^\s*$/;
	next if /^\s*#/;
	s/^\s*//;
	s/\s*$//;
	$map->{$_}=1;
    }
    close($fn);
}

sub read_2nd_column_as_flag {
    my ($map,$filename)=@_;
    open (my $fn,'<',$filename) or die "can't open $filename: $!";
    while (<$fn>) {
	chomp;
	next if /^\s*#/;
	my @in=split(/\s+/);
	next unless @in;
	$map->{$in[1]}=1;
    }
    close($fn);
}

sub read_stringval {
    my ($map,$filename)=@_;
    open (my $fn,'<',$filename) or die "$!";
    while (<$fn>) {
	chomp;
	my @in=split(/\s+/,$_,2);
	#warn 'split: in=',$_,' split: ',$in[0],'=',$in[1],"\n";
	warn "Oops! not split into 2: $_\n" if @in>2;
	$map->{$in[0]}=$in[1] if @in==2;
    }
    close($fn);
}

sub read_one2one {
    my ($map,$filename)=@_;
    open (my $fn,'<',$filename) or die "$!";
    while (<$fn>) {
	chomp;
	my @in=split(/\s+/);
	warn "Oops! not one2one: $_\n" if @in>2;
	$map->{$in[0]}=$in[1] if @in==2;
    }
    close($fn);
}

sub read_first_one2one {
    my ($map,$filename)=@_;
    open (my $fn,'<',$filename) or die "$!";
    while (<$fn>) {
	chomp;
	my @in=split(/\s+/);
	$map->{$in[0]}=$in[1] if @in>1;
	warn "read first: not one2one: $_\n" if @in>2;
    }
    close($fn);
}

sub read_one2one_with_resolver_table {
    my ($map,$filename,$resolver_table)=@_;
    open (my $fn,'<',$filename) or die "$!";
    while (<$fn>) {
	chomp;
	my @in=split(/\s+/);
	next if @in<2;
	$map->{$in[0]}=$in[1];
	if (@in>2) {
	    my $resolved=$resolver_table->{$in[0]};
	    if ($resolved) {
		$map->{$in[0]}=$resolved;
	    } else {
		warn "WARNING: not resolved ambiguous: $_\n";
	    }
	}
    }
    close($fn);
}

sub read_reverse_known_one2one {
    my ($map,$filename)=@_;
    open (my $fn,'<',$filename) or die "$!";
    while (<$fn>) {
	chomp;
	my @in=split(/\s+/);
	my $val=shift @in;
	foreach my $key (@in) {
	    $map->{$key}=$val;
	}
    }
    close($fn);
}

sub __sort_print_hasharray_fh($$$) {
    my ($fh, $map, $skip_id_map) = @_;
    foreach my $key (sort {$a cmp $b} keys(%$map)) {
	my @sorted=sort {$a cmp $b} @{$map->{$key}};
	if (@sorted==0 or not defined($sorted[0])) {
	    warn "E: empty map for $key\n";
	    next;
	}
	next if $skip_id_map and @sorted==1 and $key eq $sorted[0];
	print $fh join("\t",$key, @sorted),"\n";
    }
}

sub sort_print_hasharray {
    my ($map, $skip_id_map, $outputfile) = @_;
    my $fh = *STDOUT;
    $outputfile=undef if $outputfile and $outputfile eq '-';
    open ($fh, '>', $outputfile) or die "can't open $outputfile: $!" if defined $outputfile;
    __sort_print_hasharray_fh($fh,$map, $skip_id_map);
    close ($fh) or die "can't close $outputfile: $!" if defined $outputfile;
}

sub extract_by_subroutine {
    my ($dir, $exclude, $suffix, $func)=@_;
    foreach my $filename (glob $dir.'/*.txt') {
	next if $exclude->{$filename};
	open (my $fin,'<',$filename) or die "can't open $filename: $!";
	open (my $fno,'>',$filename.'.'.$suffix) or die "can't open $filename.$suffix: $!";
	open (my $fnew,'>',$filename.'.new') or die "can't open $filename.new: $!";
	my $comment='';
	while (<$fin>) {
	    if (/^\s*(?:#|$)/) {
		$comment .= $_;
	    } else {
		my @in=split(/\s+/, $_);
		if (&$func($filename,@in)) {
		    print $fno $comment, $_;
		} else {
		    print $fnew $comment, $_;
	    }
		$comment='';
	    }
	}
	close($fin);
	close($fno);
	close($fnew);
	if (! -s $filename.'.'.$suffix) {
	    unlink $filename.'.'.$suffix, $filename.'.new';
	}
    }
}


sub read_except_dir_distromap_map {
    &read_except_dir_distromap_with(@_[0..2],\&read_map);
}

sub read_except_dir_distromap_with {
    my ($map,$dir,$except,$func)=@_;
    foreach my $dbfile (glob $dir.'/*.txt') {
	next if $except->{$dbfile};
	print STDERR "reading dir_except $dir $dbfile...\n" if $verbose;
	&$func($map, $dbfile);
    }
}

sub get_distrodb_static_datadir {
    my ($distro)=@_;
    my $dir='/usr/share/distrodb/'.$distro.'/data';
    my $git=$ENV{'HOME'}.'/src/repo/distrodb-static-'.$distro.'.git/data';
    return $git if -d $git;
    return $dir if -d $dir;
    die "static distrodb datadir $dir not found";
}

sub read_distrodb_static_datadir_with {
    my ($map,$distro,$dbname,$func)=@_;
    foreach my $dbfile (glob &get_distrodb_static_datadir($distro).'/'.$dbname.'/*.txt') {
	print STDERR "reading static $dbfile...\n" if $verbose;
	&$func($map, $dbfile);
    }
}

sub read_distrodb_static_one2one {
    &read_distrodb_with(@_[0..2],\&read_one2one);
}


1;
