package ProjectDBTools;

use strict;
use warnings;

use RPM::Header;
use File::Basename;
use File::Path qw(make_path);
use DistroDBTools;
use Source::Repository::RPM::ALTLinuxSrcList;
use ProjectDB::Repo::RPM;
use ProjectDB::Repo::DistroDB;

use constant REPLACEABLE => -1;
use constant SUBPROJECT => -2;
use constant DIFFERENT => 0;
use constant SAME => 1;

our $verbose=0;
our $opt_warn_nourl;
our $opt_force_yes_always;
our $opt_force_yes_same_name;
our $opt_force_yes_same_name_version;
our $opt_force_yes_not_same_name;
our $opt_force_yes_not_same_version;
our $opt_force_yes_not_same_name_version;

our @LONGOPT=(
    'warn-no-url!' => \$ProjectDBTools::opt_warn_nourl,
    'force-projectdb-yes-always' => \$ProjectDBTools::opt_force_yes_always,
    'force-projectdb-yes-same-name' => \$ProjectDBTools::opt_force_yes_same_name,
    'force-projectdb-yes-same-nv' => \$ProjectDBTools::opt_force_yes_same_name_version,
    'force-projectdb-yes-same-no' => sub {},
    'force-projectdb-yes-not-same-version' => \$ProjectDBTools::opt_force_yes_not_same_version,
    'force-projectdb-yes-not-same-name' => \$ProjectDBTools::opt_force_yes_not_same_name,
    'force-projectdb-yes-not-same-nv' => \$ProjectDBTools::opt_force_yes_not_same_name_version,
    );


return 1 if $ENV{RPM_BUILD_ROOT};

my $projectdbdir_git=$ENV{PROJECTDB_GIT};
$projectdbdir_git||=$ENV{HOME}.'/src/repo/projectdb.git';
my $projectdbdir_fhs='/usr/share/projectdb';
my $projectdbdir_local='projectdb';
my $projectdbdir_external=$projectdbdir_fhs;
$projectdbdir_external=$projectdbdir_git if defined($projectdbdir_git) and -d $projectdbdir_git;

my $distromap_config=&DistroDBTools::read_distromap_config();
my $origin_projdb_suf=$distromap_config->{'ORIGINREPO'}.'-'.$distromap_config->{'ORIGINBRANCH'};
my $dest_projdb_suf=$distromap_config->{'DESTREPO'}.'-'.$distromap_config->{'DESTBRANCH'};
my $projectdb_origin2dest_filename=$origin_projdb_suf.'-'.$dest_projdb_suf.'.tsv';
my $projectdb_origin2origin_filename=$origin_projdb_suf.'-'.$origin_projdb_suf.'.tsv';
my $projectdb_dest2dest_filename=$dest_projdb_suf.'-'.$dest_projdb_suf.'.tsv';
my $originnameprefix='name://'.$distromap_config->{'ORIGINREPO'}.'/'.$distromap_config->{'ORIGINBRANCH'}.'/';
my $destnameprefix='name://'.$distromap_config->{'DESTREPO'}.'/'.$distromap_config->{'DESTBRANCH'}.'/';

my (%known_same_url, %known_replaceable_url, %known_different_url, %known_subproject_url,
    %known_same_nameuri, %known_different_nameuri, %known_replaceable_nameuri, %known_subproject_nameuri,
    %known_same_name, %known_different_name);

my ($origintoolrepo,$desttoolrepo);
my %name2rpm_cache;

sub init_project_db {
    my $origin_dict=shift;
    die "projectdb not found. Please, install projectdb\n",
	"or run with --no-projectdb option or set USE_PROJECTDB=0 in .distrodb/config\n"
	unless -d $projectdbdir_external;

    my $originrepo=Source::Repository::RPM->create_repo_chain_from_cli(
	-path=>[split(/\s+/,$distromap_config->{'ORIGIN_SRPMS'})]
	);
    my $destrepo=Source::Repository::RPM::ALTLinuxSrcList->create_repo_chain_from_cli(
	-path=>[split(/\s+/,$distromap_config->{'DEST_SRPMS'})]
	);
    $desttoolrepo=ProjectDB::Repo::RPM->new($destrepo,$destnameprefix);
    unless ($origin_dict) {
	$origintoolrepo=ProjectDB::Repo::RPM->new($originrepo,$originnameprefix);
    } else {
	$origintoolrepo=ProjectDB::Repo::DistroDB->new($originrepo,$originnameprefix,$origin_dict);
    }

    &load_map(\%known_same_url,'same',\&set_map_url);
    &load_map(\%known_different_url,'different',\&set_map_url);
    &load_map(\%known_replaceable_url,'replaceable',\&set_map_url);
    &load_map(\%known_subproject_url,'subproject',\&set_map_url);

    &load_map(\%known_same_name,'bad-origin-same',\&set_map_name);
    &load_map(\%known_same_name,'bad-dest-same',\&set_map_name);
    &load_map(\%known_different_name,'bad-origin-different',\&set_map_name);
    &load_map(\%known_different_name,'bad-dest-different',\&set_map_name);

    &load_map(\%known_same_nameuri,'nameuri-same',\&set_map_name_uri);
    &load_map(\%known_different_nameuri,'nameuri-different',\&set_map_name_uri);
    &load_map(\%known_replaceable_nameuri,'nameuri-replaceable',\&set_map_name_uri);
    &load_map(\%known_subproject_nameuri,'nameuri-subproject',\&set_map_name_uri);
}

sub load_map {
    my ($map, $dirprefix, $map_setter_sub)=@_;
    foreach my $file (&find_projectdb_files($dirprefix)) {
	print STDERR "ProjectDB: reading $file ...\n" if $verbose;
	&read_choice4($file,$map,$map_setter_sub);
    }
}

sub find_projectdb_files {
    my ($dirprefix)=@_;
    my $dirname=$dirprefix.'4';
    my @out;
    foreach my $dir (
	grep {-d $_} map {$_.'/'.$dirname}
	$projectdbdir_local, $projectdbdir_external) {
	push @out, grep {-e $_} glob($dir.'/*.tsv');
    }
    return @out;
}

sub origin2dest_same_url {
    my ($originsrcname, $destsrcname, $raw_originurl, $raw_desturl, $message, $locdebug)=@_;
    return &__same_url($origintoolrepo, $desttoolrepo, $projectdb_origin2dest_filename,
		       $originsrcname, $destsrcname, $raw_originurl, $raw_desturl, $message, $locdebug);
}

sub origin2origin_same_url {
    my ($srcname1, $srcname2, $raw_url1, $raw_url2, $message, $locdebug)=@_;
    return &__same_url($origintoolrepo, $origintoolrepo, $projectdb_origin2origin_filename,
		       $srcname1, $srcname2, $raw_url1, $raw_url2, $message, $locdebug);
}

sub dest2dest_same_url {
    my ($srcname1, $srcname2, $raw_url1, $raw_url2, $message, $locdebug)=@_;
    return &__same_url($desttoolrepo, $desttoolrepo, $projectdb_dest2dest_filename,
		       $srcname1, $srcname2, $raw_url1, $raw_url2, $message, $locdebug);
}

sub __same_url {
    my ($toolrepo1, $toolrepo2, $projectdb_filename,
	$srcname1, $srcname2, $raw_url1, $raw_url2, $message, $locdebug)=@_;

    my $nameprefix1=$toolrepo1->{-repoprefix};
    my $nameprefix2=$toolrepo2->{-repoprefix};
    my $nameuri1=$nameprefix1.$srcname1;
    my $nameuri2=$nameprefix2.$srcname2;
    return SAME if $nameuri1 eq $nameuri2;
    my $nameuritag=&__convolute($nameuri1,$nameuri2);
    return SAME if $known_same_nameuri{$nameuritag};
    return REPLACEABLE if $known_replaceable_nameuri{$nameuritag};
    return SUBPROJECT if $known_subproject_nameuri{$nameuritag};
    return 0 if $known_different_nameuri{$nameuritag};

    my $nametag=&__convolute($srcname1,$srcname2);
    return SAME if $known_same_name{$nametag};
    return 0 if $known_different_name{$nametag};

    if (! defined($raw_url1) or ! defined($raw_url2)) {
	if ($opt_warn_nourl or $locdebug) {
	    if (! defined($raw_url1) and ! defined($raw_url2)) {
		warn "NOURL: $srcname1 -> $srcname2\n",
	    } elsif (! defined($raw_url1)) {
		warn "NOORIGINURL: $srcname1 from $srcname2\t$raw_url2\n";
	    } elsif (! defined($raw_url2)) {
		warn "NODESTURL: $srcname2 from $srcname1\t$raw_url1\n";
	    }
	}
	return SAME;
    }
    my $url1=&__trim_url($raw_url1);
    my $url2=&__trim_url($raw_url2);
    if ($locdebug) {
	warn "DEBUG: same_url: trimmed originurl=$url1\n" if $url1;
	warn "DEBUG: same_url: trimmed desturl=$url2\n" if $url2;
    }
    return SAME if $url1 eq $url2;
    my $len1=length($url1);
    my $len2=length($url2);
    if ($len1<$len2 and substr($url2,0,$len1) eq $url1 or
	$len2<$len1 and substr($url1,0,$len2) eq $url2) {
	return SAME;
    }
    my $urltag=&__convolute($url1,$url2);
    return SAME if $known_same_url{$urltag};
    return REPLACEABLE if $known_replaceable_url{$urltag};
    return SUBPROJECT if $known_subproject_url{$urltag};
    return 0 if $known_different_url{$urltag};

    my $origincpan=&get_cpan_module_from_url($url1);
    my $destcpan=&get_cpan_module_from_url($url2);
    if ($locdebug) {
	warn "DEBUG: same_url: origincpan=$origincpan\n" if $origincpan;
	warn "DEBUG: same_url: destcpan=$destcpan\n" if $destcpan;
    }
    # known fedora url for perl-*; moved in import too
    return SAME if $url1 eq 'http://cpan.org' and defined($destcpan);
    return SAME if $url2 eq 'http://cpan.org' and defined($origincpan);
    if (defined($origincpan) and defined($destcpan)) {
	return SAME if $origincpan eq $destcpan;
	return 0 if !$ENV{'PROJECTDB_BROKEN_ORIGIN_CPAN'} or $origincpan ne $ENV{'PROJECTDB_BROKEN_ORIGIN_CPAN'};
    }

    my $originpypi=&get_pypi_module_from_url($url1);
    my $destpypi=&get_pypi_module_from_url($url2);
    if ($locdebug) {
	warn "DEBUG: same_url: originpypi=$originpypi\n" if $originpypi;
	warn "DEBUG: same_url: destpypi=$destpypi\n" if $destpypi;
    }
    return SAME if $url1 eq 'http://pypi.org' and defined($destpypi);
    return SAME if $url2 eq 'http://pypi.org' and defined($originpypi);
    return SAME if defined($originpypi) and defined($destpypi) and $originpypi eq $destpypi;
    if (defined($origincpan) and defined($destcpan)) {
	return SAME if $origincpan eq $destcpan;
	return 0 if !$ENV{'PROJECTDB_BROKEN_ORIGIN_PYPI'} or $origincpan ne $ENV{'PROJECTDB_BROKEN_ORIGIN_PYPI'};
    }

    # hack for alt linux kde placeholder url
    return SAME if $url2 eq 'http://kde.org' and $url1 =~m!\.kde\.org/!;

    my $toolpkg1=$toolrepo1->get_pkg($srcname1);
    if (! defined($toolpkg1)) {
	warn "W: can't find rpm for origin ",$srcname1," in ", $toolrepo1->{-repoprefix};
	return SAME;
    }
    my $toolpkg2=$toolrepo2->get_pkg($srcname2);
    if (! defined($toolpkg2)) {
	warn "W: can't find rpm for dest ",$srcname2," in ", $toolrepo2->{-repoprefix};
	return SAME;
    }

    # tmp plugs for first runs with projectdb enabled
    return SAME if $opt_force_yes_always;
    return SAME if $opt_force_yes_same_name and $srcname1 eq $srcname2;
    return SAME if $opt_force_yes_same_name_version and $srcname1 eq $srcname2 and $toolpkg1->{'VERSION'} eq $toolpkg2->{'VERSION'};
    return SAME if $opt_force_yes_not_same_name and $srcname1 ne $srcname2;
    return SAME if $opt_force_yes_not_same_version and $toolpkg1->{'VERSION'} ne $toolpkg2->{'VERSION'};
    return SAME if $opt_force_yes_not_same_name_version and ($srcname1 ne $srcname2 or $toolpkg1->{'VERSION'} ne $toolpkg2->{'VERSION'});

    #my $url1=$toolpkg1->{'URL'};
    #my $url2=$toolpkg2->{'URL'};

    $message//='';
    print STDERR "\n\n";
    print STDERR "====================================================\n";
    print STDERR '| ', $message, "\n";
    print STDERR "====================================================\n";
    print STDERR "$srcname1: $toolpkg1->{VERSION}\n";
    print STDERR "$srcname1: URL=$url1\n";
    if ($locdebug or $verbose) {
    print STDERR basename($toolpkg1->source_name),":\n";
    print STDERR "$srcname1: URL=$toolpkg1->{URL}\n";
    }
    print STDERR "$srcname1: SUMMARY=$toolpkg1->{SUMMARY}\n";
    print STDERR "$srcname1: %description:\n$toolpkg1->{DESCRIPTION}\n";
    print STDERR "----------------------------------------------------\n";
    print STDERR "$srcname2: $toolpkg2->{VERSION}\n";
    print STDERR "$srcname2: URL=$url2\n";
    if ($locdebug or $verbose) {
    print STDERR basename($toolpkg2->source_name),":\n";
    print STDERR "$srcname2: URL=$toolpkg2->{URL}\n";
    }
    print STDERR "$srcname2: SUMMARY=$toolpkg2->{SUMMARY}\n";
    print STDERR "$srcname2: %description:\n$toolpkg2->{DESCRIPTION}\n";
    print STDERR "----------------------------------------------------\n";
    my $answer;
    do {
	print STDERR "bad origin url - same(o) | different (O)/bad dest url same(d) | different (D)/
do not save: same (Y,S)/not same(N,Z)/replaceable(R)/subproject(U)/ignore (i)
same(y,s)/not same(n,z)/replaceable(r)/subproject(u)/exit(e) ?\n";
	$answer=<STDIN>;
    } until $answer =~ /^([uUoOdDynrszeYSNZRit])/;
    $answer=$1;
    exit if $answer eq 'e';
    my @write_args=($toolpkg1->{'NAME'},$toolpkg2->{'NAME'},basename($toolpkg1->source_name),basename($toolpkg2->source_name),$raw_url1,$raw_url2,$nameprefix1,$nameprefix2,$projectdb_filename);
    my $trimmedurl1=&__trim_url($url1);
    my $trimmedurl2=&__trim_url($url2);
    my $trimmedname1=&__trim_name($url1);
    my $trimmedname2=&__trim_name($url2);
    if ($answer eq 'i') {
	return SAME;
    } elsif ($answer eq 'y' or $answer eq 's') {
	&write_choice4('same',@write_args);
	$known_same_url{&__convolute($trimmedurl1,$trimmedurl2)}=1;
	$known_same_url{&__convolute($trimmedurl2,$trimmedurl1)}=1;
	return SAME;
    } elsif ($answer eq 'Y' or $answer eq 'S') {
	$known_same_url{&__convolute($trimmedurl1,$trimmedurl2)}=1;
	$known_same_url{&__convolute($trimmedurl2,$trimmedurl1)}=1;
	return SAME;
    } elsif ($answer eq 't') {
	&write_choice4('todo',@write_args);
	$known_same_url{&__convolute($trimmedurl1,$trimmedurl2)}=1;
	$known_same_url{&__convolute($trimmedurl2,$trimmedurl1)}=1;
	return SAME;
    } elsif ($answer eq 'n' or $answer eq 'z') {
	&write_choice4('different',@write_args);
	$known_different_url{&__convolute($trimmedurl1,$trimmedurl2)}=1;
	$known_different_url{&__convolute($trimmedurl2,$trimmedurl1)}=1;
	return 0;
    } elsif ($answer eq 'N' or $answer eq 'Z') {
	$known_different_url{&__convolute($trimmedurl1,$trimmedurl2)}=1;
	$known_different_url{&__convolute($trimmedurl2,$trimmedurl1)}=1;
	return 0;
    } elsif ($answer eq 'r') {
	&write_choice4('replaceable',@write_args);
	$known_replaceable_url{&__convolute($trimmedurl1,$trimmedurl2)}=1;
	$known_replaceable_url{&__convolute($trimmedurl2,$trimmedurl1)}=1;
	return REPLACEABLE;
    } elsif ($answer eq 'R') {
	$known_replaceable_url{&__convolute($trimmedurl1,$trimmedurl2)}=1;
	$known_replaceable_url{&__convolute($trimmedurl2,$trimmedurl1)}=1;
	return REPLACEABLE;
    } elsif ($answer eq 'u') {
	&write_choice4('subproject',@write_args);
	$known_subproject_url{&__convolute($trimmedurl1,$trimmedurl2)}=1;
	$known_subproject_url{&__convolute($trimmedurl2,$trimmedurl1)}=1;
	return SUBPROJECT;
    } elsif ($answer eq 'U') {
	$known_subproject_url{&__convolute($trimmedurl1,$trimmedurl2)}=1;
	$known_subproject_url{&__convolute($trimmedurl2,$trimmedurl1)}=1;
	return SUBPROJECT;
    } elsif ($answer eq 'o') {
	&write_choice4('bad-origin-same',@write_args);
	$known_same_name{&__convolute($trimmedname1,$trimmedname2)}=1;
	return SAME;
    } elsif ($answer eq 'O') {
	&write_choice4('bad-origin-different',@write_args);
	$known_different_name{&__convolute($trimmedname1,$trimmedname2)}=1;
	return 0;
    } elsif ($answer eq 'd') {
	&write_choice4('bad-dest-same',@write_args);
	$known_same_name{&__convolute($trimmedname1,$trimmedname2)}=1;
	return SAME;
    } elsif ($answer eq 'D') {
	&write_choice4('bad-dest-different',@write_args);
	$known_different_name{&__convolute($trimmedname1,$trimmedname2)}=1;
	return 0;
    }
    return SAME;
}

# convolute 2 strings into a single hash key
sub __convolute {
    return join('@@@',@_);
}

sub __trim_url {
    my $url=shift;
    $url=~s!/$!!;
    $url=~s!/index\.(htm|html|shtml|php)$!!;
    $url=~s!^https://!http://!;
    $url=~s!^http://www\.!http://!;
    $url=~s!^http://sourceforge\.(?:net|io)/projects/([^/]+)!http://$1.sourceforge.net!;
    $url=~s!^http://sf\.(?:net|io)/projects/([^/]+)!http://$1.sourceforge.net!;
    $url=~s!\.sourceforge\.io(/|$)!.sourceforge.net$1!;
    $url=~s!\.sf\.(?:net|io)(/|$)!.sourceforge.net$1!;
    $url=~s!://([^./]+)\.github\.(?:io|com)/(\S*)$!://github.com/$1/$2!;
    $url=~s!^http://hosted.fedoraproject.org/projects/(\S*)$!http://fedorahosted.org/$1!;
    $url=~s!(\.|/)pypi\.io(/|$)!$1pypi.org$2!;
    return $url;
}

sub __trim_name_uri {
    my $name=shift;
    $name=~s!/[^/]+$!!;
    return $name;
}

sub __trim_name {
    my $name=shift;
    $name=~s!/[^/]+$!!;
    $name=~s!^.*/!!;
    return $name;
}

sub write_choice4 {
    my ($choice,$srcname1,$srcname2,$srcfile1,$srcfile2,$url1,$url2,$nameprefix1,$nameprefix2,$projectdb_filename)=@_;
    my $choicedir=$projectdbdir_local.'/'.$choice.'4';
    make_path($choicedir);
    my $choicefile=$choicedir.'/'.$projectdb_filename;
    open (my $fn, '>>', $choicefile) || die "open $choicefile: $!";
    print $fn
	$nameprefix1.$srcname1.'/'.$srcfile1."\t".$url1."\t".
	$nameprefix2.$srcname2.'/'.$srcfile2."\t".$url2."\n";
    close($fn) || die "close $choicefile: $!";
}

sub read_choice4 {
    my ($choicefile,$map, $map_setter_sub)=@_;
    open (my $fn, '<', $choicefile) || die "open $choicefile: $!";
    local $_;
    while (<$fn>) {
	chomp;
	my @in=split(/\t/,$_);
	&$map_setter_sub($map,\@in);
    }
    close($fn) || die "close $choicefile: $!";
}

sub set_map_url {
    my ($map, $in)=@_;
    my $url1=&__trim_url($in->[1]);
    my $url3=&__trim_url($in->[3]);
    $map->{&__convolute($url1,$url3)}=1;
    $map->{&__convolute($url3,$url1)}=1;
}

sub set_map_name {
    my ($map, $in)=@_;
    $map->{&__convolute(&__trim_name($in->[0]),&__trim_name($in->[2]))}=1;
}

sub set_map_name_uri {
    my ($map, $in)=@_;
    my $uri1=&__trim_name_uri($in->[0]);
    my $uri2=&__trim_name_uri($in->[2]);
    $map->{&__convolute($uri1,$uri2)}=1;
    $map->{&__convolute($uri2,$uri1)}=1;
}

sub get_cpan_module_from_url {
    my $url=shift;
    my $module;
    if ($url=~m!^http://search.cpan.org/dist/([^/]+)$!) {
	$module=$1;
#DEBUG: same_url: originurl=http://search.cpan.org/dist/URI-NamespaceMap
#DEBUG: same_url: desturl=http://metacpan.org/module//URI::NamespaceMap
    } elsif ($url=~m!^http://metacpan.org/(?:release|dict|dist|module)/+([^/]+)$!) {
	$module=$1;
	$module=~s/::/-/g;
    } elsif ($url=~m!^http://cpan.org/modules/by-module/+([^/]+)$!) {
	$module=$1;
    } else {
	return undef;
    }
    # hack around broken url in Fedora's perl-Term-Animation and company
    $module=~s/^perl[_-]//;
    return $module;
}

#URL=http://pypi.python.org/pypi/grako
#URL=http://pypi.org/project//grako
sub get_pypi_module_from_url {
    my $url=shift;
    my $module;
    if ($url=~m!^http://pypi.python.org/pypi/+([^/]+)$!) {
	$module=$1;
    } elsif ($url=~m!^http://pypi.org/project/+([^/]+)$!) {
	$module=$1;
    } else {
	return undef;
    }
    return $module;
}

sub set_nameuri {
    my ($srcname1, $srcname2, $nameuri_type, $nameuri_choice, $locdebug)=@_;
    if ($nameuri_type eq 'origin2dest') {
	&_set_nameuri($srcname1,$srcname2,$origintoolrepo,$desttoolrepo,
		      $projectdb_origin2dest_filename, $nameuri_choice, $locdebug);
    } elsif ($nameuri_type eq 'origin2origin') {
	&_set_nameuri($srcname1,$srcname2,$origintoolrepo,$origintoolrepo,
		      $projectdb_origin2origin_filename, $nameuri_choice, $locdebug);
    } elsif ($nameuri_type eq 'dest2dest') {
	&_set_nameuri($srcname1,$srcname2,$desttoolrepo,$desttoolrepo,
		      $projectdb_dest2dest_filename, $nameuri_choice, $locdebug);
    } else {
	die "OOPS";
    }
}

sub _set_nameuri {
    my ($srcname1, $srcname2, $toolrepo1, $toolrepo2, $projectdb_filename, $choice, $locdebug)=@_;
    my $toolpkg1=$toolrepo1->get_pkg($srcname1);
    if (! defined($toolpkg1)) {
	die "E: can't find toolpkg for $srcname1 in ", $toolrepo1->{-repoprefix};
    }
    my $toolpkg2=$toolrepo2->get_pkg($srcname2);
    if (! defined($toolpkg2)) {
	die "E: can't find toolpkg for $srcname2 in ", $toolrepo2->{-repoprefix};
    }

    my $raw_url1=$toolpkg1->{URL};
    $raw_url1//='(none)';
    my $raw_url2=$toolpkg2->{URL};
    $raw_url2//='(none)';
    &write_choice4($choice,$toolpkg1->{'NAME'},$toolpkg2->{'NAME'},basename($toolpkg1->source_name),basename($toolpkg2->source_name),$raw_url1,$raw_url2,$toolrepo1->{-repoprefix},$toolrepo2->{-repoprefix},$projectdb_filename);
}

1;
